/*
 * ** Network and Service Differentiation Extensions to GridSim 3.0 **
 *
 * Gokul Poduval & Chen-Khong Tham
 * Computer Communication Networks (CCN) Lab
 * Dept of Electrical & Computer Engineering
 * National University of Singapore
 * August 2004
 *
 * Licence: GPL - http://www.gnu.org/copyleft/gpl.html
 *
 * Scheduler.java - Implements a scheduler
 *
 */

package gridsim.net;

import eduni.simjava.*;
import gridsim.*;
import java.util.*;


/**
 * This class provides a template for schedulers that will be used at routers.
 * Every egress port of a router needs to instantiate a PacketScheduler object
 * which it uses to determine the order in which packets should be sent out.
 *
 * @invariant $none
 * @since GridSim Toolkit 3.1
 * @author Gokul Poduval & Chen-Khong Tham, National University of Singapore
 */
public abstract class PacketScheduler
{
    private String name_ ;
    private double baudRate_;

    /**
     * Creates a new packet scheduler with the specified name and baud rate
     * (in bits/s). The name can be useful for debugging purposes, but server
     * no functional purpose.
     *
     * @param name       Name of this scheduler
     * @param baudRate   baud rate in bits/s of the port that is using
     *                   this scheduler.
     * @throws ParameterException This happens when the name is null or
     *                   the baud rate <= 0
     * @pre name != null
     * @pre baudRate > 0
     * @post $none
     */
    public PacketScheduler(String name, double baudRate) 
                           throws ParameterException
    {
        // check the entity name
        if (name == null || name.length() == 0)
        {
            String error = "PacketScheduler(): Error - name can not be " +
                           "null or empty.";
            System.out.println(error);
            throw new ParameterException(error);
        }

        // check the baud rate
        if (baudRate <= 0)
        {
            String msg = name + ": Error - the baud rate must be > 0.";
            System.out.println(msg);
            throw new ParameterException(msg);
        }

        this.name_ = name;
        this.baudRate_ = baudRate;
    }

    /**
     * Creates a new packet scheduler with the specified baud rate (bits/s).
     * The name is set to a generic <b>"PacketScheduler"</b>.
     * @param baudRate   baud rate in bits/s of the port that is using
     *                   this scheduler.
     * @throws ParameterException This happens when the baud rate <= 0
     * @pre baudRate > 0
     * @post $none
     */
    public PacketScheduler(double baudRate) throws ParameterException
    {
        this.name_ = "PacketScheduler";

        // check the baud rate
        if (baudRate <= 0)
        {
            String msg = name_ + ": Error - the baud rate must be > 0.";
            System.out.println(msg);
            throw new ParameterException(msg);
        }

        this.baudRate_ = baudRate;
    }

    /**
     * Creates a new packet scheduler with the specified name. The baud rate is
     * left at 0, and should be set with {@link #setBaudRate(double)}
     * before being the simulation is started.
     *
     * @param name Name of this scheduler
     * @throws ParameterException This happens when the name is null
     * @pre name != null
     * @post $none
     */
    public PacketScheduler(String name) throws ParameterException
    {
        // check the entity name
        if (name == null || name.length() == 0)
        {
            String error = "PacketScheduler(): Error - name can not be " +
                           "null or empty.";
            System.out.println(error);
            throw new ParameterException(error);
        }

        this.name_ = name;
        this.baudRate_ = 0.0;
    }

    /**
     * Creates a new packet scheduler with the name <b>"PacketScheduler"</b>.
     * The baud rate is left at 0, and should be set with 
     * {@link #setBaudRate(double)}
     * before being the simulation is started.
     * @pre $none
     * @post $none
     */
    public PacketScheduler()
    {
        this.name_ = "PacketScheduler";
        this.baudRate_ = 0.0;
    }


    /**
     * Returns the baud rate of the egress port that is using this scheduler.
     * If the baud rate is zero, it means you haven't set it up.
     * @return the baud rate in bits/s
     * @see gridsim.net.PacketScheduler#setBaudRate(double)
     * @pre $none
     * @post $result >= 0
     */
    public double getBaudRate() {
        return baudRate_;
    }


    /**
     * Sets the baud rate that this scheduler will be sending packets at.
     * @param rate the baud rate of this scheduler (in bits/s)
     * @return <tt>true</tt> if it is successful, <tt>false</tt> otherwise
     * @pre rate > 0
     * @post $none
     */
    public boolean setBaudRate(double rate)
    {
        if (rate < 0.0) {
            return false;
        }

        this.baudRate_ = rate;
        return true;
    }

    /**
     * This method enques a packet in this scheduler. If the implementing class
     * has buffer management policies too, then it should return <b>true</b> if the
     * packet was successfully enqued. If the packet was dropped, or could not
     * be accomodated due to any other reason, it should return <b>false</b>.
     *
     * @param np    A Packet to be enqued by this scheduler.
     * @return <tt>true</tt> if enqued, <tt>false</tt> otherwise
     * @pre np != null
     * @post $none
     */
    public abstract boolean enque(Packet np);


    /**
     * Removes a single packet and returns it. This packet should be sent out
     * by the router from the port using this scheduler.
     *
     * @return the packet to be sent out
     * @pre $none
     * @post $none
     */
    public abstract Packet deque();


    /**
     * Determines whether the scheduler is currently keeping any packets in
     * its queue(s).
     *
     * @return <tt>true</tt> if no packets are enqueued, <tt>false</tt>
     *         otherwise
     * @pre $none
     * @post $none
     */
    public abstract boolean isEmpty();

    /**
     * Determines the number of packets that are currently enqueued in this
     * scheduler.
     *
     * @return the number of packets enqueud by this scheduler.
     * @pre $none
     * @post $none
     */
    public abstract int size();

    /**
     * Returns the name of this scheduler, if one was specified during setup.
     * Otherwise "PacketScheduler" is returned. This could be used for debugging
     * purposes.
     *
     * @return the name of this scheduler
     * @pre $none
     * @post $none
     */
    public String getName() {
        return name_ ;
    }

} // end class

